from importlib.metadata import distribution
import os
import pathlib
import yaml
import json
from kubernetes_yaml_constants import *
from os import environ, name
from pathlib import Path
import argparse
from copy import deepcopy
from shutil import copyfile
from pathlib import Path
from uuid import uuid4


PROD = False
TESTCLUSTER_DEVICE_IP = None

BUILD_DIR = 'deploy'
EXTERNAL_PATH = f'{BUILD_DIR}/external.json'
INGRESS_DIR = 'ingress'
INGRESS_TEMPLATE_PATH = f'{INGRESS_DIR}/nginx-ingress-v1.1.0.yaml'
INGRESS_CONFIG_FILENAME = f'nginx.tmpl'
INGRESS_CONFIG_TEMPLATE_PATH = f'{INGRESS_DIR}/{INGRESS_CONFIG_FILENAME}'
INGRESS_CONFIG_MOUNT_PATH = "/etc/nginx/template"
INGRESS_CONFIG_VOLUME_NAME = "nginx-config" 
INGRESS_CONFIG_CUSTOM_MAX_CLIENT_BODY_SIZE_TAG = "#custom_max_client_body_size_definitions_here"
INGRESS_PROD_ISSUER_PATH = f'{INGRESS_DIR}/prod_issuer.yaml'
INGRESS_STAGING_ISSUER_PATH = f'{INGRESS_DIR}/staging_issuer.yaml'

INGRESS_DEPLOYMENT_DEPLOY_PATH = "ingress_deployment"
PROD_INGRESS_DEPLOY_FILENAME = "prod_ingress_config"
TEST_INGRESS_DEPLOY_FILENAME = "test_ingress_config"
DISTRIBUTION_INGRESS_DEPLOY_FILENAME = "distribution_ingress_config"
ENDPOINTS_FILENAME = "endpoints"
PROD_ISSUER_FILENAME = "prod_issuer"
STAGING_ISSUER_FILENAME = "staging_issuer"

DEPLOYMENTS_FIELD = "deployments"
NAMESPACE_FIELD = "namespace"
NAME_FIELD = "name"
SERVICE_ACCOUNT_FIELD = "serviceAccount"
VOLUMES_FIELD = "volumes"
URI_FIELD = "uri"
MAX_UPLOAD_SIZE_FIELD = "maxUploadSize"
PORTS_FIELD = "ports"
PORT_FIELD = "port"
TARGET_PORT_FIELD = "targetPort"
PORT_PROTOCOL_FIELD = "protocol"
EXPOSED_HTTP_PORT_FIELD = "exposedHttpPort"
CLUSTER_IP_PORTS_FIELD = "clusterIP"
LB_PORTS_FIELD = "loadBalancer"
NODE_PORT_PORTS_FIELD = "nodePort"
SIZE_FIELD = "size"
TYPE_FIELD = "type"
TYPE_PV_VALUE = "pv"
TYPE_CONFIG_MAP_VALUE = "configMap"
ACCESS_MODES_FIELD = "accessModes"
NFS_FIELD = "nfs"
NFS_SERVER_FIELD = "server"
PATH_FIELD = "path"
HOSTPATH_FIELD = "hostpath"
DEVICE_HOSTNAME_FIELD = "deviceHostname"
MOUNT_PATH_FIELD = "mountPath"
CONTENT_FIELD = "content"
SUBPATH_FIELD = "subPath"
IMAGE_FIELD = "image"
RULES_FIELD = "rules"
API_GROUPS_FIELD = "apiGroups"
RESOURCES_FIELD = "resources"
VERBS_FIELD = "verbs"
ENV_FIELD = "env"
PRIVILEGED_FIELD = "privileged"
IMAGE_PULL_SECRETS_FIELD = "imagePullSecrets"

SERVICEACCOUNT_TYPE_CLUSTER = "cluster"
SERVICEACCOUNT_TYPE_NAMESPACE = "namespace"

TYPE_NONE = "none"
TYPE_NGINX = "nginx"
TYPE_GITLAB = "gitlab"
TYPE_GITLAB_RUNNER = "gitlab-runner"
TYPE_NFS = "nfs"
TYPE_MATRIX = "matrix"
TYPE_TURN = "turn"
TYPE_WORDPRESS = "wordpress"
TYPE_NEXTCLOUD = "nextcloud"
TYPE_DOCKER_REGISTRY = "docker-registry"
TYPE_OCTOBOT = "octobot"
TYPE_SYNAPSE_ADMIN = "synapse-admin"
TYPE_DDCLIENT = "ddclient"
TYPE_MAUTRIX_BRIDGE = "mautrix-bridge"

ID_FIELD = "id"

EXTERNALS_ENDPOINTS_FIELD = "endpoints"
EXTERNALS_IP_FIELD = "ip"
EXTERNALS_PORT_FIELD = "port"
EXTERNALS_PROTOCOLS_FIELD = "protocols"
EXTERNALS_TARGETPORT_FIELD = "targetPort"
EXTERNALS_REDIRECTS_FIELD = "redirects"
EXTERNALS_ENDPOINT_FIELD = "endpoint"

PORT_DEFAULT = 80
SECRETNAME_DEFAULT = "echo-tls"
INGRESS_SECRETNAME_PROD = "prod-tls"
INGRESS_SECRETNAME_TEST = "test-tls"
INGRESS_SECRETNAME_DISTRIBUTION = "distribution-tls"

DEFAULT_INGRESS_NAME = "default-ingress"
DISTRIBUTION_INGRESS_NAME = "distribution-ingress"
TEST_INGRESS_NAME = "test-ingress"
INGRESS_CERTIFICATE_PROD = "letsencrypt-prod"
INGRESS_CERTIFICATE_STAGING = "letsencrypt-staging"

REGISTRY_420JOOS_DEV_CREDS_SECRET_NAME = "registry-420joos-dev-creds"
IMAGE_PULL_SECRETS_DEFAULT_VALUE = [REGISTRY_420JOOS_DEV_CREDS_SECRET_NAME]
IMAGE_PULL_SECRETS = {
    REGISTRY_420JOOS_DEV_CREDS_SECRET_NAME: "docker_registry/docker_registry_creds_secret.yaml",
}


def make_deployment_scripts(middle_deployment_scripts: list, middle_deployments: list):
  #get all deployment files
  nginx_bridges = []
  if middle_deployment_scripts is not None:
    for middle_deployment_script in middle_deployment_scripts:
      project_name, namespace, deployment_file_content, persistent_volumes_file_content, service_account_file_content, current_nginx_bridges = make_deployment_script_from_script(middle_deployment_script)
      if current_nginx_bridges:
        nginx_bridges.append((project_name, namespace, current_nginx_bridges))

      dump_files(project_name, deployment_file_content, persistent_volumes_file_content, service_account_file_content)

  for middle_deployment in middle_deployments:
    project_name, namespace, deployment_file_content, persistent_volumes_file_content, service_account_file_content, current_nginx_bridges = make_deployment_script_from_object(middle_deployment)
    if current_nginx_bridges:
      nginx_bridges.append((project_name, namespace, current_nginx_bridges))

    dump_files(project_name, deployment_file_content, persistent_volumes_file_content, service_account_file_content)

  prod_ingresses, test_ingresses, distribution_ingress, ingress_deployment_template, endpoints, prod_issuer, staging_issuer = make_ingresses(nginx_bridges)

  for project_name, prod_ingress in prod_ingresses:
    dump_file([prod_ingress], project_name + "_ingress")

  for project_name, test_ingress in test_ingresses:
    dump_file([test_ingress], project_name + "_testingress")

  dump_file([distribution_ingress], DISTRIBUTION_INGRESS_DEPLOY_FILENAME)
  dump_file(endpoints, ENDPOINTS_FILENAME)
  dump_file([prod_issuer], PROD_ISSUER_FILENAME)
  dump_file([staging_issuer], STAGING_ISSUER_FILENAME)
  dump_file(ingress_deployment_template, INGRESS_DEPLOYMENT_DEPLOY_PATH)


def make_deployment_script_from_script(middle_deployment_script_path: str):
  middle_deployment = {}
  with open(middle_deployment_script_path, 'r', encoding='utf-8') as f:
    middle_deployment = yaml.load(f, yaml.CLoader)

  return make_deployment_script(middle_deployment) 


def make_deployment_script_from_object(middle_deployment: dict):
  return make_deployment_script(middle_deployment)


def make_deployment_script(middle_deployment: dict):
  deployment_file_content = []
  persistent_volumes_file_content = []
  service_account_file_content = []
  nginx_bridges = []

  deployments = middle_deployment[DEPLOYMENTS_FIELD]
  ns_name = middle_deployment[NAMESPACE_FIELD]

  project_name = middle_deployment[NAME_FIELD] if middle_deployment.get(NAME_FIELD, None) is not None else ns_name

  ns = make_ns(ns_name)
  deployment_file_content.append(ns)
  persistent_volumes_file_content.append(ns)
  service_account_file_content.append(ns)

  for deployment in deployments:
    deployment_name, _, http_service, env_config_map, cluster_ip_service, lb_services, node_port_service, kubernetes_volumes, config_map_volumes, service_account, role, role_binding, uris, max_upload_size = make_deployment_components(ns_name, deployment)
    kubernetes_deployment = make_deployment(deployment_name, ns_name, deployment, kubernetes_volumes, config_map_volumes, service_account, env_config_map)

    persistent_volumes = [pv for pv, pvc, volume_name in kubernetes_volumes]
    persistent_volume_claims = [pvc for pv, pvc, volume_name in kubernetes_volumes]
    config_map_volumes = [volume for volume, volume_name in config_map_volumes]

    deployment_file_content.extend(make_deployment_file_content(http_service, env_config_map, cluster_ip_service, lb_services, node_port_service, config_map_volumes, kubernetes_deployment))
    persistent_volumes_file_content.extend(make_persistent_volumes_file_content(persistent_volumes, persistent_volume_claims))
    service_account_file_content.extend(make_service_account_file_content(service_account, role, role_binding))
    if http_service or uris != []:
      nginx_bridges.append((http_service, uris, max_upload_size))

    # image_pull_secrets ugly workaround
    image_pull_secret_names = deployment.get(IMAGE_PULL_SECRETS_FIELD, None)
    image_pull_secret_paths = []
    if image_pull_secret_names is None:
      for secret_name in IMAGE_PULL_SECRETS_DEFAULT_VALUE:
        image_pull_secret_paths.append(IMAGE_PULL_SECRETS[secret_name])
    else:
      for secret_name in image_pull_secret_names:
        image_pull_secret_paths.append(IMAGE_PULL_SECRETS[secret_name])

    image_pull_secrets = []
    for image_pull_secret_path in image_pull_secret_paths:
      secret = load_registry_secret(image_pull_secret_path)
      secret[KUBERNETES_METADATA_KEY][KUBERNETES_NAMESPACE_KEY] = ns_name
      if secret not in deployment_file_content:
        image_pull_secrets.append(secret)

    deployment_file_content.extend(image_pull_secrets)

  return project_name, ns, deployment_file_content, persistent_volumes_file_content, service_account_file_content, nginx_bridges

def make_deployment_components(ns_name, deployment: dict):
  deployment_name = deployment[NAME_FIELD]
  bridge_service = {}
  http_service = {}
  uris = []
  max_upload_size = "1m"
  env_config_map = {}
  cluster_ip_service = {}
  lb_services = []
  node_port_service = {}
  kubernetes_volumes = []
  config_map_volumes = []
  service_account = {}
  role = {}
  role_binding = {}

  if deployment.get(ENV_FIELD, None) is not None:
    env_config_map = make_env_config_map(deployment[ENV_FIELD], deployment_name, ns_name)

  if deployment.get(EXPOSED_HTTP_PORT_FIELD, None) is not None:
    bridge_service = make_bridge_service(deployment_name, ns_name, KUBERNETES_NAMESPACE_DEFAULT)
    http_service = make_exposed_cluster_ip_service(deployment_name, ns_name, deployment[EXPOSED_HTTP_PORT_FIELD])
    
    uris = deployment[URI_FIELD]
    if isinstance(uris, str):
      uris = [uris]

    if deployment.get(MAX_UPLOAD_SIZE_FIELD, None) is not None:
      max_upload_size = deployment[MAX_UPLOAD_SIZE_FIELD]

  if deployment.get(PORTS_FIELD, None) is not None:
    additional_ports = deployment[PORTS_FIELD]

    if additional_ports.get(CLUSTER_IP_PORTS_FIELD, None) is not None:
      cluster_ip_service = make_cluster_ip_service(deployment_name, ns_name, additional_ports[CLUSTER_IP_PORTS_FIELD])

    if additional_ports.get(LB_PORTS_FIELD, None) is not None:
      lb_services = make_lb_services(deployment_name, ns_name, additional_ports[LB_PORTS_FIELD])
      
    if additional_ports.get(NODE_PORT_PORTS_FIELD, None) is not None:
      node_port_service = make_node_port_service(deployment_name, ns_name, additional_ports[NODE_PORT_PORTS_FIELD])

  if deployment.get(VOLUMES_FIELD, None) is not None:
    kubernetes_volumes, config_map_volumes = make_volumes(deployment[VOLUMES_FIELD], deployment_name, ns_name)

  if deployment.get(SERVICE_ACCOUNT_FIELD, None) is not None:
    sa_type = deployment[SERVICE_ACCOUNT_FIELD].get(TYPE_FIELD, None)
    service_account, role, role_binding = make_service_account(f'{deployment_name}-sa', deployment_name, ns_name, deployment[SERVICE_ACCOUNT_FIELD][RULES_FIELD], sa_type)

  return deployment_name, bridge_service, http_service, env_config_map, cluster_ip_service, lb_services, node_port_service, kubernetes_volumes, config_map_volumes, service_account, role, role_binding, uris, max_upload_size


def make_file_content(*args):
  file_content = []
  for arg in args:
    if isinstance(arg, dict):
      if arg:
        file_content.append(arg)
    elif isinstance(arg, list):
      for item in arg:
        file_content.extend(make_file_content(item))
    else:
      raise TypeError()

  return file_content


def load_registry_secret(path: str):
  return load_yaml_file_single(path)



def dump_files(project_name: str, deployment_file_content: list, persistent_volumes_file_content: list, service_account_file_content: list):
  dump_file(deployment_file_content, project_name)
  dump_file(persistent_volumes_file_content, f'{project_name}_pvs')
  dump_file(service_account_file_content, f'{project_name}_sa')


def dump_file(file_content, file_name):
  with open(f'{file_name}.yaml', 'w', encoding='utf-8') as f:
    yaml.dump_all(file_content, f)
    

def make_ns(ns_name: str):
  ns = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_DEFAULT_VALUE,    
    KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_NAMESPACE,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAME_KEY: ns_name,
    },
  }

  return ns


def make_bridge_service(deployment_name: str, deployment_ns_name: str, service_ns_name: str = KUBERNETES_NAMESPACE_DEFAULT):
  bridge_service = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_DEFAULT_VALUE,
    KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_SERVICE,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAME_KEY: f'{deployment_ns_name}-{deployment_name}-bridge',
      KUBERNETES_NAMESPACE_KEY: service_ns_name
    },
    KUBERNETES_SPEC_KEY: {
      KUBERNETES_TYPE_KEY: KUBERNETES_SERVICE_TYPE_VALUE_EXTERNAL_NAME,
      KUBERNETES_EXTERNALNAME_KEY: f"{deployment_name}.{deployment_ns_name}.svc.cluster.local"
    },
  }

  return bridge_service


def make_service(deployment_name: str, ns_name: str, type: str, ports):
  name = deployment_name
  if type == KUBERNETES_SERVICE_TYPE_VALUE_LOADBALANCER:
    name = f"{deployment_name}-{ports[0][PORT_PROTOCOL_FIELD].lower()}-{ports[0][PORT_FIELD]}" # lb service only has one port per service

  service = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_DEFAULT_VALUE,
    KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_SERVICE,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAME_KEY: name,
      KUBERNETES_NAMESPACE_KEY: ns_name
    },
    KUBERNETES_SPEC_KEY: {
      KUBERNETES_TYPE_KEY: type,
      KUBERNETES_SELECTOR_KEY: {
        KUBERNETES_APP_KEY: deployment_name,
      },
      KUBERNETES_PORTS_KEY: make_ports(ports),
    },
  }

  return service


def make_ports(ports):
  kubernetes_ports = []

  for port_config in ports:
    kubernetes_port = {}

    if PORT_FIELD in port_config:
      kubernetes_port[KUBERNETES_PORT_KEY] = port_config[PORT_FIELD]

    if TARGET_PORT_FIELD in port_config:
      kubernetes_port[KUBERNETES_TARGETPORT_KEY] = port_config[TARGET_PORT_FIELD]

    if PORT_PROTOCOL_FIELD in port_config:
      kubernetes_port[KUBERNETES_PROTOCOL_KEY] = port_config[PORT_PROTOCOL_FIELD]

    if NAME_FIELD in port_config:
      kubernetes_port[KUBERNETES_NAME_KEY] = port_config[NAME_FIELD]
    else:
      port_name = ''

      if KUBERNETES_PROTOCOL_KEY in kubernetes_port.keys():
        port_name += kubernetes_port[KUBERNETES_PROTOCOL_KEY].lower()
      else:
        port_name += KUBERNETES_PROTOCOL_VALUE_TCP
      
      if KUBERNETES_PORT_KEY in kubernetes_port.keys():
        port_name += f'-{kubernetes_port[KUBERNETES_PORT_KEY]}'

      kubernetes_port[KUBERNETES_NAME_KEY] = port_name
    
    kubernetes_ports.append(kubernetes_port)

  return kubernetes_ports


def make_persistent_volume(volume_name:str, volume_specs: dict, deployment_name: str, ns: str=KUBERNETES_NAMESPACE_DEFAULT):
  accessModes = volume_specs.get(ACCESS_MODES_FIELD) if volume_specs.get(ACCESS_MODES_FIELD, None) is not None else [KUBERNETES_ACCESSMODES_VALUE_READ_WRITE_MANY,]

  volume_name = f"{ns}-{deployment_name}-{volume_name}"

  specs = {
    KUBERNETES_STORAGECLASSNAME_KEY: KUBERNETES_STORAGECLASSNAME_DEFAULT_VALUE,
    KUBERNETES_CAPACITY_KEY: {
      KUBERNETES_STORAGE_KEY: volume_specs[SIZE_FIELD],
    },
    KUBERNETES_PERSISTENT_VOLUME_RECLAIM_POLICY_KEY: KUBERNETES_PERSISTENT_VOLUME_RECLAIM_POLICY_VALUE_RETAIN,
    KUBERNETES_ACCESSMODES_KEY: accessModes,
  }

  if volume_specs.get(NFS_FIELD, None) is not None:
    server = volume_specs[NFS_FIELD][NFS_SERVER_FIELD] if PROD else TESTCLUSTER_DEVICE_IP
    server = os.path.expandvars(server)

    specs.update({
      KUBERNETES_NFS_KEY: {
        KUBERNETES_NFS_SERVER_KEY: server,
        KUBERNETES_PATH_KEY: volume_specs[NFS_FIELD][PATH_FIELD],
      },
      KUBERNETES_NFS_MOUNT_OPTIONS_KEY: KUBERNETES_NFS_MOUNT_OPTIONS_DEFAULT_VALUE,
    })

  elif volume_specs.get(HOSTPATH_FIELD, None) is not None:
    specs.update({
      KUBERNETES_HOSTPATH_KEY: {
        KUBERNETES_PATH_KEY: volume_specs[HOSTPATH_FIELD][PATH_FIELD],
      },
      KUBERNETES_NODE_AFFINITY_KEY: {
        KUBERNETES_REQUIRED_KEY: {
          KUBERNETES_NODE_SELECTOR_TERMS_KEY: [
            {
              KUBERNETES_MATCH_EXPRESSIONS_KEY: [
                {
                  KUBERNETES_KEY_KEY: KUBERNETES_LABEL_HOSTNAME_VALUE,
                  KUBERNETES_OPERATOR_KEY: KUBERNETES_OPERATOR_IN_VALUE,
                  KUBERNETES_VALUES_KEY: [volume_specs[HOSTPATH_FIELD][DEVICE_HOSTNAME_FIELD]],
                },
              ],
            },
          ],
        },
      },
    })


  persistent_volume = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_DEFAULT_VALUE,
    KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_PERSISTENTVOLUME,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAME_KEY: volume_name,
      KUBERNETES_NAMESPACE_KEY: KUBERNETES_NAMESPACE_DEFAULT,
      KUBERNETES_LABELS_KEY: {
        KUBERNETES_APP_KEY: deployment_name,
      },
    },
    KUBERNETES_SPEC_KEY: specs
  }

  return persistent_volume


def make_persistent_volume_claim(volume_name:str, volume_specs: dict, pv: dict, deployment_name: str, ns_name: str):
  accessModes = volume_specs.get(ACCESS_MODES_FIELD) if volume_specs.get(ACCESS_MODES_FIELD, None) is not None else [KUBERNETES_ACCESSMODES_VALUE_READ_WRITE_MANY]

  persistent_volume_claim = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_DEFAULT_VALUE,
    KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_PERSISTENTVOLUMECLAIM,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAME_KEY: f"{deployment_name}-{volume_name}",
      KUBERNETES_NAMESPACE_KEY: ns_name,
      KUBERNETES_LABELS_KEY: {
        KUBERNETES_APP_KEY: deployment_name,
      },
    },
    KUBERNETES_SPEC_KEY: {
      KUBERNETES_VOLUMENAME_KEY: pv[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY],
      KUBERNETES_STORAGECLASSNAME_KEY: KUBERNETES_STORAGECLASSNAME_DEFAULT_VALUE,
      KUBERNETES_ACCESSMODES_KEY: accessModes,
      KUBERNETES_RESOURCES_KEY: {
        KUBERNETES_REQUESTS_KEY: {
          KUBERNETES_STORAGE_KEY: volume_specs[SIZE_FIELD]
        },
      },
    },
  }

  return persistent_volume_claim


def make_volumes(volumes: dict, deployment_name: str, ns_name: str):
  pvs = []
  config_maps = []

  for volume_name, volume_specs in volumes.items():
    if volume_specs[TYPE_FIELD] == TYPE_PV_VALUE:
      pv = make_persistent_volume(volume_name, volume_specs, deployment_name, ns_name)
      pvc = make_persistent_volume_claim(volume_name, volume_specs, pv, deployment_name, ns_name)
      # pvs.append(pv)
      # pvcs.append(pvc)
      pvs.append((pv, pvc, volume_name))

    elif volume_specs[TYPE_FIELD] == TYPE_CONFIG_MAP_VALUE:
      config_map = make_config_map(volume_name, volume_specs, ns_name, deployment_name)
      config_maps.append((config_map, volume_name))

  return pvs, config_maps


def make_config_map(name: str, specs: dict, ns_name: str, deployment_name: str):
  config_map_raw_content = {}

  if isinstance(specs[CONTENT_FIELD], dict): # case collection of variables
    config_map_raw_content = specs[CONTENT_FIELD]
  elif isinstance(specs[CONTENT_FIELD], list): # case collection of files
    for file in specs[CONTENT_FIELD]:
      filename = Path(file).name
      try:
        # open as unicode
        with open(file, 'r', encoding='utf-8') as f:
          config_map_raw_content[filename] = f.read()
      except UnicodeDecodeError:
        # open as binary
        with open(file, 'rb') as f:
          config_map_raw_content[filename] = f.read()

  config_map_content = {}
  for key, value in config_map_raw_content.items():
    value_with_env_vars = os.path.expandvars(value)
    config_map_content.update({key: value_with_env_vars})

  config_map = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_DEFAULT_VALUE,
    KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_CONFIG_MAP,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAME_KEY: f"{deployment_name}-{name}",
      KUBERNETES_NAMESPACE_KEY: ns_name,
    },
    KUBERNETES_DATA_KEY: config_map_content
  }

  return config_map


# def make_service_accounts(service_accounts: dict, deployment_name: str, ns_name: str):
#   kubernetes_service_accounts = []

#   for service_account_name, service_account_specs in service_accounts.items():
#     service_account, role, role_binding = make_service_account(service_account_name, deployment_name, ns_name, service_account_specs[RULES_FIELD], service_account_specs[TYPE_FIELD])
#     kubernetes_service_accounts.append((service_account, role, role_binding))

#   return kubernetes_service_accounts


def make_service_account(name: str, deployment_name: str, ns_name: str, rules: list, type: str):
  kubernetes_rules = []

  for rule in rules:
    rule_config = {
      KUBERNETES_SERVICEACCOUNT_APIGROUPS_KEY: rule[API_GROUPS_FIELD],
      KUBERNETES_RESOURCES_KEY: rule[RESOURCES_FIELD],
      KUBERNETES_SERVICEACCOUNT_VERBS_KEY: rule[VERBS_FIELD],
    }

    kubernetes_rules.append(rule_config)

  service_account = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_DEFAULT_VALUE,
    KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_SERVICEACCOUNT,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAME_KEY: name,
      KUBERNETES_NAMESPACE_KEY: ns_name,
      KUBERNETES_LABELS_KEY: {
        KUBERNETES_APP_KEY: deployment_name
      },
    },
  }

  role = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_RBAC_AUTHORIZATION_DEFAULT_VALUE,
    KUBERNETES_SERVICEACCOUNT_RULES_KEY: kubernetes_rules,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAME_KEY: name,
      KUBERNETES_LABELS_KEY: {
        KUBERNETES_APP_KEY: deployment_name
      },
    },
  }

  if type == SERVICEACCOUNT_TYPE_CLUSTER:
    role.update({KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_CLUSTERROLE})
    role[KUBERNETES_METADATA_KEY].update({KUBERNETES_NAMESPACE_KEY: KUBERNETES_NAMESPACE_DEFAULT})
  else:
    role.update({KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_ROLE})
    role[KUBERNETES_METADATA_KEY].update({KUBERNETES_NAMESPACE_KEY: ns_name})

  role_binding = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_RBAC_AUTHORIZATION_DEFAULT_VALUE,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAME_KEY: name,
      KUBERNETES_LABELS_KEY: {
        KUBERNETES_APP_KEY: deployment_name
      },
    },
    KUBERNETES_SERVICEACCOUNT_ROLEREF_KEY: {
      KUBERNETES_SERVICEACCOUNT_APIGROUP_KEY: KUBERNETES_API_VERSION_RBAC_AUTHORIZATION_VALUE,
      KUBERNETES_KIND_KEY: role[KUBERNETES_KIND_KEY],
      KUBERNETES_NAME_KEY: role[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY],
    },
    KUBERNETES_SERVICEACCOUNT_SUBJECTS_KEY: [
      {
        KUBERNETES_KIND_KEY: service_account[KUBERNETES_KIND_KEY],
        KUBERNETES_NAME_KEY: service_account[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY],
        KUBERNETES_NAMESPACE_KEY: service_account[KUBERNETES_METADATA_KEY][KUBERNETES_NAMESPACE_KEY],
      },
    ],
  }

  if type == SERVICEACCOUNT_TYPE_CLUSTER:
    role_binding.update({KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_CLUSTERROLEBINDING})
    role_binding[KUBERNETES_METADATA_KEY].update({KUBERNETES_NAMESPACE_KEY: KUBERNETES_NAMESPACE_DEFAULT})
  elif type == SERVICEACCOUNT_TYPE_NAMESPACE:
    role_binding.update({KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_ROLEBINDING})
    role_binding[KUBERNETES_METADATA_KEY].update({KUBERNETES_NAMESPACE_KEY: ns_name})

  return service_account, role, role_binding


def make_deployment(name: str, ns_name: str, deployment_config: dict, kubernetes_volumes: list, config_map_volumes: list, service_account: dict, env: dict):
  deployment_volumes, volume_mounts = make_deployment_volumes_volume_mounts(kubernetes_volumes, config_map_volumes, deployment_config.get(VOLUMES_FIELD, None))
  open_container_ports = []

  priviledged = deployment_config.get(PRIVILEGED_FIELD, None)
  if priviledged is None:
    priviledged = KUBERNETES_PRIVILEGED_DEFAULT_VALUE

  if deployment_config.get(EXPOSED_HTTP_PORT_FIELD, None) is not None:
    open_container_ports.append({KUBERNETES_DEPLOYMENT_CONTAINERPORT_KEY: deployment_config[EXPOSED_HTTP_PORT_FIELD]})

  ports = deployment_config.get(PORTS_FIELD, None)
  if ports is not None:
    for port_type in ports.values():
      if port_type is not None:
        for port_config in port_type:
          container_port = port_config[TARGET_PORT_FIELD]
          if container_port not in [open_container_port[KUBERNETES_DEPLOYMENT_CONTAINERPORT_KEY] for open_container_port in open_container_ports]:
            open_container_ports.append({KUBERNETES_DEPLOYMENT_CONTAINERPORT_KEY: container_port})

  env_from = [
    {
      KUBERNETES_DEPLOYMENT_CONFIGMAPREF_KEY: {
        KUBERNETES_NAME_KEY: env[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY],
      },
    },
  ] if env else []

  # note about IgnoredDuringExecution: https://www.magalix.com/blog/influencing-kubernetes-scheduler-decisions, but it's not important right now
  affinity = {
    KUBERNETES_NODE_AFFINITY_KEY: {
      KUBERNETES_REQUIRED_DURING_SCHEDULING_IGNORED_DURING_EXECUTION_KEY:{
        KUBERNETES_NODE_SELECTOR_TERMS_KEY : [
          {
            KUBERNETES_MATCH_EXPRESSIONS_KEY: [
              {
                KUBERNETES_KEY_KEY: KUBERNETES_LABEL_NODE_ROLE_MASTER_VALUE,
                KUBERNETES_OPERATOR_KEY: KUBERNETES_OPERATOR_DOES_NOT_EXIST_VALUE,
              },
            ],
          },
        ],
      },
    },
  } if not any([KUBERNETES_NODE_AFFINITY_KEY in pv[KUBERNETES_SPEC_KEY] or KUBERNETES_NODE_AFFINITY_KEY in pvc[KUBERNETES_SPEC_KEY] for pv, pvc, volume_name in kubernetes_volumes]) else {}

  image_pull_secrets = deployment_config.get(IMAGE_PULL_SECRETS_FIELD, None)
  if image_pull_secrets is None:
    image_pull_secrets = IMAGE_PULL_SECRETS_DEFAULT_VALUE

  image_pull_secrets = [{KUBERNETES_NAME_KEY: image_pull_secret} for image_pull_secret in image_pull_secrets]
  
  deployment = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_NAMESPACE_DEFAULT_VALUE,
    KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_DEPLOYMENT,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAME_KEY: name,
      KUBERNETES_NAMESPACE_KEY: ns_name,
    },
    KUBERNETES_SPEC_KEY: {
      KUBERNETES_DEPLOYMENT_REPLICAS_KEY: KUBERNETES_DEPLOYMENT_REPLICAS_DEFAULT_VALUE,
      KUBERNETES_SELECTOR_KEY: {
        KUBERNETES_MATCHLABELS_KEY: {
          KUBERNETES_APP_KEY: name,
        },
      },
      KUBERNETES_DEPLOYMENT_TEMPLATE_KEY: {
        KUBERNETES_METADATA_KEY: {
          KUBERNETES_LABELS_KEY: {
            KUBERNETES_APP_KEY: name,
          },
        },
        KUBERNETES_SPEC_KEY: {
          KUBERNETES_DEPLOYMENT_SERVICEACCOUNTNAME_KEY: service_account[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY] if service_account else "",
          KUBERNETES_DEPLOYMENT_CONTAINERS_KEY: [
            {
              KUBERNETES_NAME_KEY: name,
              KUBERNETES_DEPLOYMENT_IMAGE_KEY: os.path.expandvars(deployment_config[IMAGE_FIELD]),
              KUBERNETES_DEPLOYMENT_IMAGEPULLPOLICY_KEY: KUBERNETES_DEPLOYMENT_IMAGEPULLPOLICY_DEFAULT_VALUE,
              KUBERNETES_PORTS_KEY: open_container_ports,
              KUBERNETES_DEPLOYMENT_VOLUMEMOUNTS_KEY: volume_mounts,
              KUBERNETES_DEPLOYMENT_ENVFROM_KEY: env_from,
              KUBERNETES_DEPLOYMENT_SECURITYCONTEXT_KEY: {
                KUBERNETES_PRIVILEGED_KEY: priviledged,
              },
            },
          ],
          KUBERNETES_DEPLOYMENT_IMAGE_PULL_SECRETS_KEY: image_pull_secrets,
          KUBERNETES_DEPLOYMENT_VOLUMES_KEY: deployment_volumes,
          KUBERNETES_AFFINITY_KEY: affinity,
        },
      },
    },
  }

  return deployment


def make_deployment_volumes_volume_mounts(kubernetes_volumes: list, config_maps: list, volumes_config: dict):
  deployment_volumes = []
  volume_mounts = []

  for pv, pvc, volume_name in kubernetes_volumes:
    deployment_volume, volume_mount = make_deployment_pvc_volume_mount(volume_name, pvc, volumes_config[volume_name][MOUNT_PATH_FIELD])

    deployment_volumes.append(deployment_volume)
    volume_mounts.append(volume_mount)

  for config_map, volume_name in config_maps:
    configmap_volumes, configmap_volume_mounts = make_deployment_configmap_volume_mount(volume_name, config_map, volumes_config[volume_name][MOUNT_PATH_FIELD], volumes_config[volume_name].get(SUBPATH_FIELD, False))

    deployment_volumes.extend(configmap_volumes)
    volume_mounts.extend(configmap_volume_mounts)

  # for volume_name, volume_config in volumes.items():
  #   mount_path = volume_config[MOUNT_PATH_FIELD]
  #   volume_type = volume_config[TYPE_FIELD]
  #   volume_type_key = ''

  #   if volume_type == TYPE_PV_VALUE:
  #     volume_type_key = KUBERNETES_DEPLOYMENT_PERSISTENTVOLUMECLAIM_KEY
    
  #   if volume_type == TYPE_CONFIG_MAP_VALUE:
  #     volume_type_key = KUBERNETES_DEPLOYMENT_CONFIG_MAP_KEY

  #   deployment_volume, volume_mount = make_deployment_volume_volume_mount(volume_name, volume_type_key, mount_path)
    
  #   deployment_volumes.append(deployment_volume)
  #   volume_mounts.append(volume_mount)

  return deployment_volumes, volume_mounts


def make_deployment_pvc_volume_mount(volume_name: str, pvc: dict, mount_path: str):
  pvc_name = pvc[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY]
  
  volume = {
    KUBERNETES_NAME_KEY: volume_name,
    KUBERNETES_DEPLOYMENT_PERSISTENTVOLUMECLAIM_KEY: {
      KUBERNETES_DEPLOYMENT_CLAIMNAME_KEY: pvc_name,
    },
  }

  volume_mount = make_deployment_volume_mount(volume_name, mount_path)

  return volume, volume_mount

def make_deployment_configmap_volume_mount(volume_name: str, config_map: dict, mount_path: str, sub_pathed: bool):
  volumes = []
  volume_mounts = []

  config_map_name = config_map[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY]
  
  if sub_pathed:
    for filename in config_map[KUBERNETES_DATA_KEY].keys():
      subpath_volume_name = f"{volume_name}-{filename.replace('.', '-')}"
      volume = make_deployment_configmap_volume(subpath_volume_name, config_map_name)
      volume[KUBERNETES_DEPLOYMENT_CONFIG_MAP_KEY].update({
        KUBERNETES_ITEMS_KEY: [
          {
            KUBERNETES_KEY_KEY: filename,
            KUBERNETES_PATH_KEY: filename,
          },
        ],
      })

      volume_mount = make_deployment_volume_mount(subpath_volume_name, mount_path, filename)

      volumes.append(volume)
      volume_mounts.append(volume_mount)
  else:
    volume = make_deployment_configmap_volume(volume_name, config_map_name)
    volume_mount = make_deployment_volume_mount(volume_name, mount_path)

    volumes.append(volume)
    volume_mounts.append(volume_mount)

  return volumes, volume_mounts

def make_deployment_configmap_volume(volume_name: str, config_map_name: str):
  return {
    KUBERNETES_NAME_KEY: volume_name,
    KUBERNETES_DEPLOYMENT_CONFIG_MAP_KEY: {
      KUBERNETES_NAME_KEY: config_map_name,
    },
  }

def make_deployment_volume_mount(volume_name: str, mount_path: str, sub_path: str=None):
  volume_mount = {
    KUBERNETES_NAME_KEY: volume_name,
  }

  if sub_path is not None:
    mount_path = str(pathlib.Path(mount_path) / pathlib.Path(sub_path))
    volume_mount.update({
      KUBERNETES_DEPLOYMENT_MOUNTPATH_KEY: mount_path,
      KUBERNETES_SUBPATH_KEY: sub_path,
    })
  else:
    volume_mount.update({
      KUBERNETES_DEPLOYMENT_MOUNTPATH_KEY: mount_path,
    })

  return volume_mount
  

def make_exposed_cluster_ip_service(deployment_name: str, ns_name: str, exposed_http_port):
  tcp_port = {
    PORT_FIELD: PORT_DEFAULT,
    TARGET_PORT_FIELD: exposed_http_port,
    PORT_PROTOCOL_FIELD: KUBERNETES_PROTOCOL_VALUE_TCP
  }

  return make_cluster_ip_service(deployment_name, ns_name, [tcp_port])


def make_endpoint(name: str, ip: str, middle_ports_config: list):
  ports = make_ports(middle_ports_config)

  endpoint = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_DEFAULT_VALUE,
    KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_ENDPOINTS,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAMESPACE_KEY: KUBERNETES_NAMESPACE_DEFAULT,
      KUBERNETES_NAME_KEY: name,
      # KUBERNETES_LABELS_KEY: {
      #   KUBERNETES_APP_KEY: name,
      # },
    },
    KUBERNETES_ENDPOINTS_SUBSETS_KEY: [
      {
        KUBERNETES_ENDPOINTS_ADDRESSES_KEY: [
          {
            KUBERNETES_ENDPOINTS_IP_KEY: ip,
          },
        ],
        KUBERNETES_PORTS_KEY: ports,
      },
    ],
  }

  return endpoint


def make_endpoint_service(name: str, middle_port_configs: list):
  endpoint_service = make_service(name, KUBERNETES_NAMESPACE_DEFAULT, KUBERNETES_SERVICE_TYPE_VALUE_CLUSTER_IP, middle_port_configs)
  endpoint_service[KUBERNETES_SPEC_KEY][KUBERNETES_SERVICE_SPEC_CLUSTERIP_KEY] = KUBERNETES_SERVICE_SPEC_CLUSTERIP_NONE_VALUE
  endpoint_service[KUBERNETES_SPEC_KEY].pop(KUBERNETES_SELECTOR_KEY)

  return endpoint_service

def import_from_file(path: str):
  with open(path, 'r', encoding='utf-8') as f:
    return list(yaml.load_all(f, yaml.CLoader))


def make_cluster_ip_service(deployment_name: str, ns_name: str, ports):
  return make_service(deployment_name, ns_name, KUBERNETES_SERVICE_TYPE_VALUE_CLUSTER_IP, ports)


def make_lb_services(deployment_name: str, ns_name: str, ports):
  # for each port, a new service has to be created, occurs only in loadbalancer
  lb_services = []

  for port in ports:
    lb_service = make_service(deployment_name, ns_name, KUBERNETES_SERVICE_TYPE_VALUE_LOADBALANCER, [port])
    lb_services.append(lb_service)
  return lb_services
  

def make_node_port_service(deployment_name: str, ns_name: str, ports):
  return make_service(deployment_name, ns_name, KUBERNETES_SERVICE_TYPE_VALUE_NODE_PORT, ports)


def make_env_config_map(specs: dict, deployment_name: str, ns_name: str):
  return make_config_map('env', specs, ns_name, deployment_name)

def make_deployment_file_content(http_service: dict, env: dict, cluster_ip_service: dict, 
                                  lb_services: list, node_port_service: list, config_map_volumes: list, deployment: dict):

  return make_file_content(http_service, env, cluster_ip_service, lb_services, node_port_service, config_map_volumes, deployment)


def make_persistent_volumes_file_content(pvs: list, pvcs: list):
  return make_file_content(pvs, pvcs)


def make_service_account_file_content(service_account: dict, role: dict, role_binding: dict):
  return make_file_content(service_account, role, role_binding)



def publish_external_source(path: str):
  content = import_from_file(path)
  #publish content


def load_nginx_middle_template():
  return load_yaml_file_single('deploy/templates/nginx/nginx_middle_template.yaml')

def load_gitlab_middle_template():
  return load_yaml_file_single('deploy/templates/gitlab/gitlab_middle_template.yaml')

def load_nfs_middle_template():
  return load_yaml_file_single('deploy/templates/nfs/nfs_middle_template.yaml')

def load_matrix_middle_template():
  return load_yaml_file_single('deploy/templates/synapse/synapse_middle_template.yaml')

def load_turn_middle_template():
  return load_yaml_file_single('deploy/templates/turn/turn_middle_template.yaml')

def load_wordpress_middle_template():
  return load_yaml_file_single('deploy/templates/wordpress/wordpress_middle_template.yaml')

def load_nextcloud_middle_template():
  return load_yaml_file_single('deploy/templates/nextcloud/nextcloud_middle_template.yaml')

def load_gitlab_runner_middle_template():
  return load_yaml_file_single('deploy/templates/gitlab_runner/gitlab_runner_middle_template.yaml')

def load_docker_registry_middle_template():
  return load_yaml_file_single('deploy/templates/docker_registry/docker_registry_middle_template.yaml')

def load_octobot_middle_template():
  return load_yaml_file_single('deploy/templates/octobot/octobot_middle_template.yaml')

def load_synapse_admin_middle_template():
  return load_yaml_file_single('deploy/templates/synapse_admin/synapse_admin_middle_template.yaml')

def load_ddclient_middle_template():
  return load_yaml_file_single('deploy/templates/ddclient/ddclient_middle_template.yaml')

def load_mautrix_bridge_middle_template():
  return load_yaml_file_single('deploy/templates/synapse/bridges/mautrix_bridge_middle_template.yaml')


def load_yaml_file_single(path: str):
  yaml_content = {}
  with open(path, 'r') as f:
    yaml_content = yaml.load(f, yaml.CLoader)

  return yaml_content


def small_to_middle_deployments(small_deployments: list):
  middle_deployments = []

  if small_deployments is not None:
    for small_deployment_path in small_deployments:
      small_deployment = {}
      with open(small_deployment_path, 'r') as f:
        small_deployment = yaml.load(f, yaml.CLoader)

      middle_deployment_template = {}
      if small_deployment[TYPE_FIELD] == TYPE_NGINX:
        middle_deployment_template = load_nginx_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_GITLAB:
        middle_deployment_template = load_gitlab_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_NFS:
        middle_deployment_template = load_nfs_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_MATRIX:
        middle_deployment_template = load_matrix_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_TURN:
        middle_deployment_template = load_turn_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_WORDPRESS:
        middle_deployment_template = load_wordpress_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_NEXTCLOUD: 
        middle_deployment_template = load_nextcloud_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_GITLAB_RUNNER: 
        middle_deployment_template = load_gitlab_runner_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_DOCKER_REGISTRY: 
        middle_deployment_template = load_docker_registry_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_OCTOBOT: 
        middle_deployment_template = load_octobot_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_SYNAPSE_ADMIN: 
        middle_deployment_template = load_synapse_admin_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_DDCLIENT: 
        middle_deployment_template = load_ddclient_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_MAUTRIX_BRIDGE: 
        middle_deployment_template = load_mautrix_bridge_middle_template()
      elif small_deployment[TYPE_FIELD] == TYPE_NONE:
        middle_deployment_template = small_deployment
      else:
        print(f"no template for {small_deployment[TYPE_FIELD]} found. skipping")

      if middle_deployment_template:
        middle_deployment = deepmerge_dicts(middle_deployment_template, small_deployment)
        middle_deployment.pop(TYPE_FIELD)
        middle_deployments.append(middle_deployment)

  return middle_deployments


def deepmerge_dicts(template: dict, update_values: dict):
  merged = {}
  merged.update(template)

  for key, value in update_values.items():
    merged_node = {}
    node_template = merged.get(key, None)
    if isinstance(value, dict):
      if node_template is None:
        merged_node = {key: value}
      else:
        merged_node = {key: deepmerge_dicts(node_template, value)}
    elif isinstance(value, list):
      if node_template is None:
        merged_node = {key: value}
      else:
        merged_list = []
        for node_value in value:
          if isinstance(node_value, str):
            merged_list.append(node_value)

          elif isinstance(node_value, dict):
            item_id = node_value.get(ID_FIELD, None)
            if item_id is not None: # case container_definition
              inner_node_template = next((item for item in node_template if item.get(ID_FIELD, None) is not None and item[ID_FIELD] == item_id), None)

              if inner_node_template is None:
                merged_list.append(node_value)
              else:
                merged_list.append(deepmerge_dicts(inner_node_template, node_value))
            else:
              merged_list.append(node_value)

          else:
            raise ValueError(f"Cannot merge type {type(node_value)} into the list")          

        merged_node = {key: merged_list}
    else:
      merged_node = {key: value}
    
    merged.update(merged_node)

  return merged


def make_endpoints(endpoint_configs: dict):
  endpoints = []

  for endpoint_name, endpoint_config in endpoint_configs.items():
    endpoint_middle_port_configs = []
    endpoint_service_middle_port_configs = []
    port_configs = endpoint_config[PORTS_FIELD]
    max_upload_size = endpoint_config[MAX_UPLOAD_SIZE_FIELD]

    for port_number, port_config in port_configs.items():
      port_number = int(port_number)
      for protocol in port_config[EXTERNALS_PROTOCOLS_FIELD]:
        endpoint_port = {
          PORT_FIELD: port_number,
          PORT_PROTOCOL_FIELD: protocol,
        }

        endpoint_service_port = {
          PORT_FIELD: port_number,
          PORT_PROTOCOL_FIELD: protocol,
        }

        endpoint_middle_port_configs.append(endpoint_port)
        endpoint_service_middle_port_configs.append(endpoint_service_port)

    endpoint = make_endpoint(endpoint_name, endpoint_config[EXTERNALS_IP_FIELD], endpoint_middle_port_configs)
    endpoint_service = make_endpoint_service(endpoint_name, endpoint_service_middle_port_configs)

    endpoints.append((endpoint, endpoint_service, max_upload_size))

  return endpoints


def make_test_endpoint():
  endpoints = {
    'testcluster-endpoint': {
      EXTERNALS_IP_FIELD: TESTCLUSTER_DEVICE_IP,
      MAX_UPLOAD_SIZE_FIELD: '1024m',
      PORTS_FIELD: {
        '443': {
          EXTERNALS_PROTOCOLS_FIELD: ["TCP",],
        },
      }
    }
  }
  return make_endpoints(endpoints)[0]


def make_ingress_rule(uri: str, service_name: str, service_port: int):
  rule = {
    KUBERNETES_INGRESS_HOST_KEY: uri,
    KUBERNETES_INGRESS_HTTP_KEY: {
      KUBERNETES_INGRESS_PATHS_KEY: [
        {
          KUBERNETES_PATH_KEY: KUBERNETES_INGRESS_PATH_DEFAULT_VALUE,
          KUBERNETES_INGRESS_PATHTYPE_KEY: KUBERNETES_INGRESS_PATHTYPE_DEFAULT_VALUE,
          KUBERNETES_INGRESS_BACKEND_KEY: {
            KUBERNETES_INGRESS_SERVICE_KEY: {
              KUBERNETES_NAME_KEY: service_name,
              KUBERNETES_PORT_KEY: {
                KUBERNETES_INGRESS_SERVICEPORTNUMBER_KEY: service_port,
              },
            },
          },
        },
      ],
    },
  }

  return rule


def make_ingress(name: str, hosts: list, rules: list, certificate_name: str, secretName: str = SECRETNAME_DEFAULT, redirect_https: bool=False, namespace=KUBERNETES_NAMESPACE_DEFAULT):
  annotations = {
    "cert-manager.io/cluster-issuer": certificate_name,
    "kubernetes.io/ingress.class": "nginx",
  }

  annotations["nginx.ingress.kubernetes.io/backend-protocol"] = "HTTPS" if redirect_https else "HTTP"

  specs = {KUBERNETES_INGRESS_RULES_KEY: rules}
  # if not redirect_https:
  specs.update({
    KUBERNETES_INGRESS_TLS_KEY: [
      {
        KUBERNETES_INGRESS_HOSTS_KEY: hosts,
        KUBERNETES_INGRESS_SECRETNAME_KEY: secretName,
      },
    ]
  })

  ingress = {
    KUBERNETES_API_VERSION_KEY: KUBERNETES_API_VERSION_INGRESS_CONFIG_DEFAULT_VALUE,
    KUBERNETES_KIND_KEY: KUBERNETES_KIND_VALUE_INGRESS,
    KUBERNETES_METADATA_KEY: {
      KUBERNETES_NAMESPACE_KEY: namespace,
      KUBERNETES_NAME_KEY: name,
      KUBERNETES_ANNOTATIONS_KEY: annotations,
    },
    KUBERNETES_SPEC_KEY: specs,
  }

  return ingress


def make_upload_size_config(backends: list):
  header = '                # Custom upload sizes\n'
  upload_size_config = ''
  for backend, max_upload_size in backends:
    is_first_entry = len(upload_size_config) == 0
    upload_size_config += ('                {{ ' + ('if' if is_first_entry else 'else if') + ' (contains $location.Backend "' + backend + '") }}\n' + 
                          f'                client_max_body_size                    {max_upload_size};\n')

  is_first_entry = len(upload_size_config) == 0

  if not is_first_entry:
    upload_size_config += '                {{ else }}\n'
  upload_size_config += '                client_max_body_size                    1m;\n'
  if not is_first_entry:
    upload_size_config += '                {{ end }}\n'

  return header + upload_size_config


def make_ingress_entry_rules(uri: str, endpoint_name: str, service_port: int, create_test_entries=True):
  uri = os.path.expandvars(uri)
  
  test_uri = make_testenv_uri(uri) if create_test_entries else None

  prod_rule = make_ingress_rule(uri, endpoint_name, service_port)
  test_rule = make_ingress_rule(test_uri, endpoint_name, service_port) if create_test_entries else None

  return (prod_rule, uri), (test_rule, test_uri)


def make_prod_test_ingress(ingress_bridges: list):
  prod_configs = []
  test_configs = []
  upload_size_configs = []
  for project_name, namespace, ingress_config in ingress_bridges:
    ns_name = namespace[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY]
    prod_rules = []
    prod_hosts = []
    test_hosts = []
    test_rules = []
    for http_service, uris, max_upload_size in ingress_config:
      service_name = http_service[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY]
      service_port = http_service[KUBERNETES_SPEC_KEY][KUBERNETES_PORTS_KEY][0][KUBERNETES_PORT_KEY]    

      for uri in uris:
        (prod_rule, prod_uri), (test_rule, test_uri) = make_ingress_entry_rules(uri, service_name, service_port)
        prod_hosts.append(prod_uri)
        prod_rules.append(prod_rule)

        test_hosts.append(test_uri)
        test_rules.append(test_rule)

    prod_configs.append((project_name, ns_name, prod_hosts, prod_rules))
    test_configs.append((project_name, ns_name, test_hosts, test_rules))

    upload_size_configs.append((service_name, max_upload_size))

  return prod_configs, test_configs, upload_size_configs

def make_distribution_ingress(externals: dict, external_endpoints: dict, test_hosts: list, test_rules: list, test_service: dict, test_max_upload_size: str):
  external_hosts = []
  external_rules = []
  upload_size_configs = {}

  for test_rule in test_rules:
    distribution_paths = []
    for path in test_rule[KUBERNETES_INGRESS_HTTP_KEY][KUBERNETES_INGRESS_PATHS_KEY]:
      endpoint_ports = test_service[KUBERNETES_SPEC_KEY][KUBERNETES_PORTS_KEY]
      distribution_path = deepcopy(path)

      backend = distribution_path[KUBERNETES_INGRESS_BACKEND_KEY]
      backend[KUBERNETES_INGRESS_SERVICE_KEY][KUBERNETES_NAME_KEY] = test_service[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY]
      backend[KUBERNETES_INGRESS_SERVICE_KEY][KUBERNETES_PORT_KEY][KUBERNETES_INGRESS_SERVICEPORTNUMBER_KEY] = test_service[KUBERNETES_SPEC_KEY][KUBERNETES_PORTS_KEY][0][KUBERNETES_PORT_KEY]
      # backend[KUBERNETES_INGRESS_SERVICEPORT_KEY] = next((port for port in endpoint_ports if next((other_protocol_port for other_protocol_port in endpoint_ports if port[KUBERNETES_PORT_KEY] == other_protocol_port[KUBERNETES_PORT_KEY] and port[KUBERNETES_PROTOCOL_KEY] != other_protocol_port[KUBERNETES_PROTOCOL_KEY]), None) is not None), None)[KUBERNETES_PORT_KEY]

      distribution_path[KUBERNETES_INGRESS_BACKEND_KEY] = backend
      distribution_paths.append(distribution_path)
    
    test_distribution_rule = deepcopy(test_rule)
    test_distribution_rule[KUBERNETES_INGRESS_HTTP_KEY][KUBERNETES_INGRESS_PATHS_KEY] = distribution_paths
    external_rules.append(test_distribution_rule)
  
  external_hosts.extend(test_hosts)
  upload_size_configs.update({test_service[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY]: test_max_upload_size})

  for external_uri, external_config in externals.items():
    endpoint_name = external_config[EXTERNALS_ENDPOINT_FIELD]
    endpoint_port = external_config.get(PORT_FIELD, None)
    if endpoint_port is None:
      endpoint_port = int(list(external_endpoints[endpoint_name][PORTS_FIELD].keys())[0])
    max_upload_size = external_endpoints[endpoint_name][MAX_UPLOAD_SIZE_FIELD]
    (external_rule, uri), (test_rule, test_uri) = make_ingress_entry_rules(external_uri, endpoint_name, endpoint_port, create_test_entries=False)

    upload_size_configs.update({endpoint_name: max_upload_size})

    external_hosts.append(uri)
    external_rules.append(external_rule) 

  return external_hosts, external_rules, upload_size_configs.items()


def make_ingresses(ingress_bridges: list):
  prod_ingresses = []
  test_ingresses = []
  externals = {}
  with open(EXTERNAL_PATH, 'r', encoding='utf-8') as f:
    externals = json.load(f)

  external_endpoint_configs = externals[EXTERNALS_ENDPOINTS_FIELD]
  external_endpoints = make_endpoints(external_endpoint_configs)

  test_endpoint, test_service, test_max_upload_size = make_test_endpoint()

  endpoints = [test_endpoint, test_service]

  for external_endpoint_config in external_endpoints:
    external_endpoint, external_service, external_max_upload_size = external_endpoint_config
    endpoints.extend([external_endpoint, external_service])

  prod_configs, test_configs, prod_test_upload_size_configs = make_prod_test_ingress(ingress_bridges)
  test_hosts = []
  test_rules = []
  for _, _, hosts, rules in test_configs:
    test_hosts.extend(hosts)
    test_rules.extend(rules)
  
  distribution_hosts, distribution_rules, distribution_upload_size_configs = make_distribution_ingress(externals[EXTERNALS_REDIRECTS_FIELD], externals[EXTERNALS_ENDPOINTS_FIELD],  test_hosts, test_rules, test_service, test_max_upload_size)
  
  upload_size_configs = prod_test_upload_size_configs
  upload_size_configs.extend(distribution_upload_size_configs)

  for project_name, project_namespace, prod_hosts, prod_rules in prod_configs:
    prod_ingress = make_ingress(project_name, prod_hosts, prod_rules, INGRESS_CERTIFICATE_PROD, INGRESS_SECRETNAME_PROD, redirect_https=False, namespace=project_namespace)
    prod_ingresses.append((project_name, prod_ingress))

  for project_name, project_namespace, test_hosts, test_rules in test_configs:
    test_ingress = make_ingress(project_name, test_hosts, test_rules, INGRESS_CERTIFICATE_STAGING, INGRESS_SECRETNAME_TEST, redirect_https=False, namespace=project_namespace)
    test_ingresses.append((project_name, test_ingress))

  # test_ingress = make_ingress(TEST_INGRESS_NAME, test_hosts, test_rules, INGRESS_CERTIFICATE_STAGING, INGRESS_SECRETNAME_TEST, redirect_https=False) # when here prod, https should work / test, distribution ingress separate
  distribution_ingress = make_ingress(DISTRIBUTION_INGRESS_NAME, distribution_hosts, distribution_rules, INGRESS_CERTIFICATE_PROD, INGRESS_SECRETNAME_DISTRIBUTION, redirect_https=True)

  upload_size_config = make_upload_size_config(upload_size_configs)

  ingress_template_config_content = []
  with open(INGRESS_CONFIG_TEMPLATE_PATH, 'r', encoding='utf-8') as f:
    ingress_template_config_content = f.readlines()
  
  insert_index = next(i for i, line in enumerate(ingress_template_config_content) if INGRESS_CONFIG_CUSTOM_MAX_CLIENT_BODY_SIZE_TAG in line) + 1
  ingress_template_config_content.insert(insert_index, upload_size_config)
  ingress_template_config_content = ''.join(ingress_template_config_content)

  ingress_deployment_content = import_from_file(INGRESS_TEMPLATE_PATH)
  ingress_controller_deployment = next(ingress_resource for ingress_resource in ingress_deployment_content if ingress_resource.get(KUBERNETES_KIND_KEY) == KUBERNETES_KIND_VALUE_DEPLOYMENT)
  nginx_config_specs = {
    INGRESS_CONFIG_VOLUME_NAME: {
      CONTENT_FIELD: {INGRESS_CONFIG_FILENAME: ingress_template_config_content},
      MOUNT_PATH_FIELD: INGRESS_CONFIG_MOUNT_PATH,
      TYPE_FIELD: TYPE_CONFIG_MAP_VALUE,
    }
  }
  nginx_config_config_map = make_config_map(INGRESS_CONFIG_VOLUME_NAME, 
    nginx_config_specs[INGRESS_CONFIG_VOLUME_NAME], 
    ingress_controller_deployment[KUBERNETES_METADATA_KEY][KUBERNETES_NAMESPACE_KEY],
    ingress_controller_deployment[KUBERNETES_METADATA_KEY][KUBERNETES_NAME_KEY])

  deployment_volumes, volume_mounts = make_deployment_volumes_volume_mounts(
    [], 
    [(nginx_config_config_map, INGRESS_CONFIG_VOLUME_NAME)],
    nginx_config_specs)
  
  ingress_controller_deployment[KUBERNETES_SPEC_KEY][KUBERNETES_DEPLOYMENT_TEMPLATE_KEY][KUBERNETES_SPEC_KEY][KUBERNETES_DEPLOYMENT_VOLUMES_KEY].extend(deployment_volumes)
  controller_deployment = next(container for container in ingress_controller_deployment[KUBERNETES_SPEC_KEY][KUBERNETES_DEPLOYMENT_TEMPLATE_KEY][KUBERNETES_SPEC_KEY][KUBERNETES_DEPLOYMENT_CONTAINERS_KEY] if container.get(KUBERNETES_NAME_KEY) == "controller")
  controller_deployment[KUBERNETES_DEPLOYMENT_VOLUMEMOUNTS_KEY].extend(volume_mounts)

  ingress_controller_config_map = next(ingress_resource for ingress_resource in ingress_deployment_content if ingress_resource.get(KUBERNETES_KIND_KEY) == KUBERNETES_KIND_VALUE_CONFIG_MAP and ingress_resource[KUBERNETES_METADATA_KEY].get(KUBERNETES_NAME_KEY) == "ingress-nginx-controller")
  ingress_controller_config_map[KUBERNETES_DATA_KEY][KUBERNETES_INGRESS_USE_PROXY_PROTOCOL_KEY] = KUBERNETES_INGRESS_USE_PROXY_PROTOCOL_DEFAULT_VALUE

  prod_issuer = load_yaml_file_single(INGRESS_PROD_ISSUER_PATH)
  staging_issuer = load_yaml_file_single(INGRESS_STAGING_ISSUER_PATH)

  ingress_deployment_content.append(nginx_config_config_map)

  return prod_ingresses, test_ingresses, distribution_ingress, ingress_deployment_content, endpoints, prod_issuer, staging_issuer


def make_testenv_uri(original_uri: str):
    test_uri = original_uri.split('.')
    test_uri.insert(-2, 'testenv')
    return '.'.join(test_uri)


def publish_sideloads(sideload_paths: list):
  if sideload_paths is not None:
    for path in sideload_paths:
      sideloaded = Path(path)
      copyfile(sideloaded, sideloaded.name)
  return None


def parse_args():
  parser = argparse.ArgumentParser()
  parser.add_argument('--small', type=str, nargs='+')
  parser.add_argument('--middle', type=str, nargs='+')
  parser.add_argument('--sideload', type=str, nargs='+')

  return parser.parse_args()


if __name__ == "__main__":
  PROD = environ.get('PROD') == "true"
  TESTCLUSTER_DEVICE_IP = environ.get('TESTCLUSTER_DEVICE_IP')
  # generate_ingress_files()
  # test()
  args = parse_args()
  created_middle_deployments = small_to_middle_deployments(args.small)
  make_deployment_scripts(args.middle, created_middle_deployments)
  publish_sideloads(args.sideload)
  # with open("testing/test-backend.yaml", 'r', encoding='utf-8') as f:
  #   with open("testing/test-backend.json", 'w', encoding='utf-8') as ff:
  #     json.dump(yaml.load_all(f, yaml.CLoader), ff, indent=2)


# configIn = {
#   "kind": "EncryptionConfig",
#   "apiVersion": "v1",
#   "resources": [
#     {
#       "resources": [
#           "secrets"
#       ],
#       "providers": [
#         {
#           "aescbc": {
#             "keys": [
#               {
#                 "name": "key1",
#                 "secret": "%s" % (generateSecret(32))
#               }
#             ]
#           }
#         }
#       ]
#     }
#   ]
# }
